/**
 * GeoFIS Native Library
 *
 * Copyright (C) 2013 IRSTEA INRA SupAgro
 *
 * IDDN.FR.001.390016.000.R.P.2013.000.31235
 *
 * Auteurs/Authors:
 *     Jean-luc Lablée - IRSTEA
 *     Serge Guillaume - IRSTEA
 *     Brigitte Charnomordic - INRA
 *     Anne Tireau - INRA
 *     Bruno Tisseyre - SupAgro
 *
 * =========================================================================
 *
 * Ce logiciel est régi par la licence CeCILL soumise au droit français et
 * respectant les principes de diffusion des logiciels libres. Vous pouvez
 * utiliser, modifier et/ou redistribuer ce programme sous les conditions
 * de la licence CeCILL telle que diffusée par le CEA, le CNRS et l'INRIA
 * sur le site "http://www.cecill.info".
 *
 * En contrepartie de l'accessibilité au code source et des droits de copie,
 * de modification et de redistribution accordés par cette licence, il n'est
 * offert aux utilisateurs qu'une garantie limitée.  Pour les mêmes raisons,
 * seule une responsabilité restreinte pèse sur l'auteur du programme,  le
 * titulaire des droits patrimoniaux et les concédants successifs.
 *
 * A cet égard  l'attention de l'utilisateur est attirée sur les risques
 * associés au chargement,  à l'utilisation,  à la modification et/ou au
 * développement et à la reproduction du logiciel par l'utilisateur étant
 * donné sa spécificité de logiciel libre, qui peut le rendre complexe à
 * manipuler et qui le réserve donc à des développeurs et des professionnels
 * avertis possédant  des  connaissances  informatiques approfondies.  Les
 * utilisateurs sont donc invités à charger  et  tester  l'adéquation  du
 * logiciel à leurs besoins dans des conditions permettant d'assurer la
 * sécurité de leurs systèmes et ou de leurs données et, plus généralement,
 * à l'utiliser et l'exploiter dans les mêmes conditions de sécurité.
 *
 * Le fait que vous puissiez accéder à cet en-tête signifie que vous avez
 * pris connaissance de la licence CeCILL, et que vous en avez accepté les
 * termes.
 *
 * =========================================================================
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
#include <geofis/algorithm/feature/feature_normalization.hpp>
#include <boost/assign/list_of.hpp>
#include <boost/assign/assign_reference_iterator_traits.hpp>
#include <boost/range/any_range.hpp>
#include <util/functional/unary_reference_adaptor.hpp>
#include <util/gmock/matchers.hpp>
#include <gmock/gmock.h>

using namespace util;
using namespace geofis;
using namespace boost;
using namespace boost::assign;
using namespace testing;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

typedef any_range<double, single_pass_traversal_tag, const double> double_range_type;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

namespace {

struct monodimensional_feature_mock {

	typedef double attribute_type;

	MOCK_CONST_METHOD0( get_attribute, double () );
	MOCK_METHOD1( set_normalized_attribute, void (double) );
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

struct multidimensional_feature_mock {

	typedef double attribute_type;
	typedef double_range_type const_attribute_range_type;

	MOCK_CONST_METHOD0( get_attribute_range, const_attribute_range_type () );
	MOCK_METHOD1( set_normalized_attribute_range, void (const double_range_type &) );
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

struct monodimensional_function_mock {

	typedef void result_type;

	MOCK_CONST_METHOD1( unary_operator, void (double) );

	void operator()(double attribute) const {
		unary_operator(attribute);
	}
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

struct multidimensional_function_mock {

	typedef void result_type;

	MOCK_CONST_METHOD1( unary_operator, void (const double_range_type &) );

	void operator()(const double_range_type &attributes) const {
		unary_operator(attributes);
	}
};

} // namespace

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

namespace geofis {

template <> struct is_monodimensional_feature<monodimensional_feature_mock> : boost::true_type {};

template <> struct is_monodimensional_feature<const monodimensional_feature_mock> : boost::true_type {};

} // namespace geofis

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

typedef feature_normalization<monodimensional_feature_mock> monodimensional_feature_normalization_type;
typedef feature_normalization<multidimensional_feature_mock> multidimensional_feature_normalization_type;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

TEST(feature_normalization_tester, test_monodimensional_feature_normalization) {

	monodimensional_feature_mock feature_mock1, feature_mock2, feature_mock3;

	EXPECT_CALL( feature_mock1, get_attribute() ).WillRepeatedly(Return(4.0));
	EXPECT_CALL( feature_mock2, get_attribute() ).WillRepeatedly(Return(2.0));
	EXPECT_CALL( feature_mock3, get_attribute() ).WillRepeatedly(Return(3.0));

	monodimensional_feature_normalization_type normalization(monodimensional_feature_normalization_type::initialize(cref_list_of<3>(feature_mock1)(feature_mock2)(feature_mock3)));

	EXPECT_CALL( feature_mock1, set_normalized_attribute(1.0) );
	EXPECT_CALL( feature_mock2, set_normalized_attribute(0.0) );
	EXPECT_CALL( feature_mock3, set_normalized_attribute(0.5) );

	normalization.normalize(ref_list_of<3>(feature_mock1)(feature_mock2)(feature_mock3));
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

TEST(feature_normalization_tester, test_monodimensional_feature_normalization_to_function) {

	monodimensional_feature_mock feature_mock1, feature_mock2, feature_mock3;

	EXPECT_CALL( feature_mock1, get_attribute() ).WillRepeatedly(Return(4.0));
	EXPECT_CALL( feature_mock2, get_attribute() ).WillRepeatedly(Return(2.0));
	EXPECT_CALL( feature_mock3, get_attribute() ).WillRepeatedly(Return(3.0));

	monodimensional_feature_normalization_type normalization(monodimensional_feature_normalization_type::initialize(cref_list_of<3>(feature_mock1)(feature_mock2)(feature_mock3)));

	// set_normalized_attribute never called
	EXPECT_CALL( feature_mock1, set_normalized_attribute(_) ).Times(0);
	EXPECT_CALL( feature_mock2, set_normalized_attribute(_) ).Times(0);
	EXPECT_CALL( feature_mock3, set_normalized_attribute(_) ).Times(0);

	monodimensional_function_mock monodimensional_function_mock;
	InSequence normalize_order;
	EXPECT_CALL( monodimensional_function_mock, unary_operator(1.0) );
	EXPECT_CALL( monodimensional_function_mock, unary_operator(0.0) );
	EXPECT_CALL( monodimensional_function_mock, unary_operator(0.5) );

	normalization.normalize_to_function(ref_list_of<3>(feature_mock1)(feature_mock2)(feature_mock3), make_unary_reference_adaptor(monodimensional_function_mock));
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

TEST(feature_normalization_tester, test_multidimensional_feature_normalization) {

	multidimensional_feature_mock feature_mock1, feature_mock2, feature_mock3;

	EXPECT_CALL( feature_mock1, get_attribute_range() ).WillRepeatedly(Return(list_of(1.0)(2.0)(8.0)(12.0)(-2.0)(-6.0)));
	EXPECT_CALL( feature_mock2, get_attribute_range() ).WillRepeatedly(Return(list_of(1.5)(4.0)(7.0)(14.0)(-6.0)(-8.0)));
	EXPECT_CALL( feature_mock3, get_attribute_range() ).WillRepeatedly(Return(list_of(2.0)(3.0)(9.0)(10.0)(-4.0)(-10.0)));

	multidimensional_feature_normalization_type normalization(multidimensional_feature_normalization_type::initialize(cref_list_of<3>(feature_mock1)(feature_mock2)(feature_mock3)));

	EXPECT_CALL( feature_mock1, set_normalized_attribute_range(EqRange(list_of(0.0)(0.0)(0.5)(0.5)(1.0)(1.0))) );
	EXPECT_CALL( feature_mock2, set_normalized_attribute_range(EqRange(list_of(0.5)(1.0)(0.0)(1.0)(0.0)(0.5))) );
	EXPECT_CALL( feature_mock3, set_normalized_attribute_range(EqRange(list_of(1.0)(0.5)(1.0)(0.0)(0.5)(0.0))) );

	normalization.normalize(ref_list_of<3>(feature_mock1)(feature_mock2)(feature_mock3));
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

TEST(feature_normalization_tester, test_multidimensional_feature_normalization_to_function) {

	multidimensional_feature_mock feature_mock1, feature_mock2, feature_mock3;

	EXPECT_CALL( feature_mock1, get_attribute_range() ).WillRepeatedly(Return(list_of(1.0)(2.0)(8.0)(12.0)(-2.0)(-6.0)));
	EXPECT_CALL( feature_mock2, get_attribute_range() ).WillRepeatedly(Return(list_of(1.5)(4.0)(7.0)(14.0)(-6.0)(-8.0)));
	EXPECT_CALL( feature_mock3, get_attribute_range() ).WillRepeatedly(Return(list_of(2.0)(3.0)(9.0)(10.0)(-4.0)(-10.0)));

	multidimensional_feature_normalization_type normalization(multidimensional_feature_normalization_type::initialize(cref_list_of<3>(feature_mock1)(feature_mock2)(feature_mock3)));

	// set_normalized_attribute_range never called
	EXPECT_CALL( feature_mock1, set_normalized_attribute_range(_) ).Times(0);
	EXPECT_CALL( feature_mock2, set_normalized_attribute_range(_) ).Times(0);
	EXPECT_CALL( feature_mock3, set_normalized_attribute_range(_) ).Times(0);

	multidimensional_function_mock multidimensional_function_mock;
	InSequence normalize_order;
	EXPECT_CALL( multidimensional_function_mock, unary_operator(EqRange(list_of(0.0)(0.0)(0.5)(0.5)(1.0)(1.0))) );
	EXPECT_CALL( multidimensional_function_mock, unary_operator(EqRange(list_of(0.5)(1.0)(0.0)(1.0)(0.0)(0.5))) );
	EXPECT_CALL( multidimensional_function_mock, unary_operator(EqRange(list_of(1.0)(0.5)(1.0)(0.0)(0.5)(0.0))) );

	normalization.normalize_to_function(ref_list_of<3>(feature_mock1)(feature_mock2)(feature_mock3), make_unary_reference_adaptor(multidimensional_function_mock));
}
