/**
 * GeoFIS Native Library
 *
 * Copyright (C) 2013 IRSTEA INRA SupAgro
 *
 * IDDN.FR.001.390016.000.R.P.2013.000.31235
 *
 * Auteurs/Authors:
 *     Jean-luc Lablée - IRSTEA
 *     Serge Guillaume - IRSTEA
 *     Brigitte Charnomordic - INRA
 *     Anne Tireau - INRA
 *     Bruno Tisseyre - SupAgro
 *
 * =========================================================================
 *
 * Ce logiciel est régi par la licence CeCILL soumise au droit français et
 * respectant les principes de diffusion des logiciels libres. Vous pouvez
 * utiliser, modifier et/ou redistribuer ce programme sous les conditions
 * de la licence CeCILL telle que diffusée par le CEA, le CNRS et l'INRIA
 * sur le site "http://www.cecill.info".
 *
 * En contrepartie de l'accessibilité au code source et des droits de copie,
 * de modification et de redistribution accordés par cette licence, il n'est
 * offert aux utilisateurs qu'une garantie limitée.  Pour les mêmes raisons,
 * seule une responsabilité restreinte pèse sur l'auteur du programme,  le
 * titulaire des droits patrimoniaux et les concédants successifs.
 *
 * A cet égard  l'attention de l'utilisateur est attirée sur les risques
 * associés au chargement,  à l'utilisation,  à la modification et/ou au
 * développement et à la reproduction du logiciel par l'utilisateur étant
 * donné sa spécificité de logiciel libre, qui peut le rendre complexe à
 * manipuler et qui le réserve donc à des développeurs et des professionnels
 * avertis possédant  des  connaissances  informatiques approfondies.  Les
 * utilisateurs sont donc invités à charger  et  tester  l'adéquation  du
 * logiciel à leurs besoins dans des conditions permettant d'assurer la
 * sécurité de leurs systèmes et ou de leurs données et, plus généralement,
 * à l'utiliser et l'exploiter dans les mêmes conditions de sécurité.
 *
 * Le fait que vous puissiez accéder à cet en-tête signifie que vous avez
 * pris connaissance de la licence CeCILL, et que vous en avez accepté les
 * termes.
 *
 * =========================================================================
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
#include <geofis/algorithm/zoning/fusion/zone/zone.hpp>
#include <geofis/algorithm/zoning/fusion/zone/zone_area_computer.hpp>
#include <geofis/geometry/polygon_with_holes.hpp>
#include <boost/assign/list_of.hpp>
#include <boost/range/any_range.hpp>
#include <boost/range/adaptor/transformed.hpp>
#include <boost/utility/addressof.hpp>
#include <util/gmock/matchers.hpp>
#include <CGAL/Cartesian.h>
#include <gmock/gmock.h>
#include <cmath>

using namespace geofis;
using namespace CGAL;
using namespace boost;
using namespace boost::assign;
using namespace boost::adaptors;
using namespace testing;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

typedef Cartesian<double> kernel_type;
typedef Point_2<kernel_type> point_type;
typedef Polygon_2<kernel_type> polygon_type;
typedef Polygon_with_holes_2<kernel_type> polygon_with_holes_type;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

namespace {

struct feature_mock {};

struct voronoi_zone_mock {

	typedef int id_type;
	typedef feature_mock feature_type;
	typedef polygon_type geometry_type;
	typedef any_range<double, forward_traversal_tag, const double> attribute_range_type;

	MOCK_CONST_METHOD0( get_id, id_type () );
	MOCK_CONST_METHOD0( get_geometry, geometry_type () );
	MOCK_CONST_METHOD0( get_attribute_range, attribute_range_type () );

	friend bool operator==(const voronoi_zone_mock &lhs, const voronoi_zone_mock &rhs) {
	    return addressof(lhs) == addressof(rhs);
	}

	friend bool operator!=(const voronoi_zone_mock &lhs, const voronoi_zone_mock &rhs) {
	    return !(lhs == rhs);
	}
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

struct sqrt_computer {

	typedef double result_type;

	double operator()(double value) const {
		return sqrt(value);
	}
};

} // namespace

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

typedef geofis::zone<polygon_with_holes_type, voronoi_zone_mock> zone_type;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

TEST(zone_tester, test_initialize) {

	// initialize voronoi zones
	voronoi_zone_mock voronoi_zone_mock1, voronoi_zone_mock2, voronoi_zone_mock3;
	EXPECT_CALL( voronoi_zone_mock1, get_id() ).WillRepeatedly(Return(2));
	EXPECT_CALL( voronoi_zone_mock2, get_id() ).WillRepeatedly(Return(1));
	EXPECT_CALL( voronoi_zone_mock3, get_id() ).WillRepeatedly(Return(3));

	{
		zone_type zone(voronoi_zone_mock1);
		ASSERT_EQ( (size_t)1, zone.size() );
		ASSERT_EQ( 2, zone.get_id() );
		ASSERT_THAT( zone.get_voronoi_zones(), EqRange(cref_list_of<1>(voronoi_zone_mock1)) );
	}

	{
		zone_type zone(cref_list_of<3>(voronoi_zone_mock1)(voronoi_zone_mock2)(voronoi_zone_mock3));
		ASSERT_EQ( (size_t)3, zone.size() );
		ASSERT_EQ( 1, zone.get_id() );
		ASSERT_THAT( zone.get_voronoi_zones(), EqRange(cref_list_of<3>(voronoi_zone_mock1)(voronoi_zone_mock2)(voronoi_zone_mock3)) );
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

TEST(zone_tester, test_area) {

	// initialize voronoi zones
	NiceMock<voronoi_zone_mock> voronoi_zone_mock1, voronoi_zone_mock2, voronoi_zone_mock3;
	ON_CALL( voronoi_zone_mock1, get_id() ).WillByDefault(Return(2));
	polygon_type geometry1 = make_polygon<kernel_type>(list_of(point_type(0, 0))(point_type(1, 0))(point_type(1, 1))(point_type(0, 1)));
	EXPECT_CALL( voronoi_zone_mock1, get_geometry() ).WillRepeatedly(Return(geometry1));
	ON_CALL( voronoi_zone_mock2, get_id() ).WillByDefault(Return(1));
	polygon_type geometry2 = make_polygon<kernel_type>(list_of(point_type(1, 0))(point_type(2, 0))(point_type(2, 2))(point_type(1, 2)));
	EXPECT_CALL( voronoi_zone_mock2, get_geometry() ).WillRepeatedly(Return(geometry2));
	ON_CALL( voronoi_zone_mock3, get_id() ).WillByDefault(Return(3));
	polygon_type geometry3 = make_polygon<kernel_type>(list_of(point_type(2, 0))(point_type(4, 0))(point_type(4, 2))(point_type(2, 2)));
	EXPECT_CALL( voronoi_zone_mock3, get_geometry() ).WillRepeatedly(Return(geometry3));

	{
		zone_type zone((const voronoi_zone_mock &)voronoi_zone_mock1);
		ASSERT_EQ( 1.0, zone.get_area() );
	}

	{
		zone_type zone(cref_list_of<3>(voronoi_zone_mock1)(voronoi_zone_mock2)(voronoi_zone_mock3));
		ASSERT_EQ( 7.0, zone.get_area() );
	}

	{ // zone area optimization
		zone_type zone(cref_list_of<1>(voronoi_zone_mock1));
		zone_type zone1((const voronoi_zone_mock &)voronoi_zone_mock2);
		zone_type zone2((const voronoi_zone_mock &)voronoi_zone_mock3);
		compute_zone_area(zone, zone1, zone2);
		ASSERT_EQ( 6.0, zone.get_area() );
	}

}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

TEST(zone_tester, test_geometry) {

	// initialize voronoi zones
	NiceMock<voronoi_zone_mock> voronoi_zone_mock1;
	polygon_type geometry1 = make_polygon<kernel_type>(list_of(point_type(0, 0))(point_type(1, 0))(point_type(1, 1))(point_type(0, 1)));
	ON_CALL( voronoi_zone_mock1, get_id() ).WillByDefault(Return(0));
	EXPECT_CALL( voronoi_zone_mock1, get_geometry() ).WillOnce(Return(geometry1));

	NiceMock<voronoi_zone_mock> voronoi_zone_mock2;
	polygon_type geometry2 = make_polygon<kernel_type>(list_of(point_type(1, 0))(point_type(2, 0))(point_type(2, 1))(point_type(1, 1)));
	ON_CALL( voronoi_zone_mock2, get_id() ).WillByDefault(Return(0));
	EXPECT_CALL( voronoi_zone_mock2, get_geometry() ).WillOnce(Return(geometry2));

	// initialize zone
	zone_type zone(cref_list_of<2>(voronoi_zone_mock1)(voronoi_zone_mock2));

	ASSERT_EQ( make_polygon_with_holes<kernel_type>(list_of(point_type(0, 0))(point_type(1, 0))(point_type(2, 0))(point_type(2, 1))(point_type(1, 1))(point_type(0, 1))), zone.get_geometry() );
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

TEST(zone_tester, test_attributes) {

	// initialize voronoi zones
	NiceMock<voronoi_zone_mock> voronoi_zone_mock1;
	ON_CALL( voronoi_zone_mock1, get_id() ).WillByDefault(Return(0));
    EXPECT_CALL( voronoi_zone_mock1, get_attribute_range() ).WillOnce(Return(list_of(1.0)(10.0)));

    NiceMock<voronoi_zone_mock> voronoi_zone_mock2;
	ON_CALL( voronoi_zone_mock2, get_id() ).WillByDefault(Return(0));
    EXPECT_CALL( voronoi_zone_mock2, get_attribute_range() ).WillOnce(Return(list_of(2.0)(20.0)));

    NiceMock<voronoi_zone_mock> voronoi_zone_mock3;
	ON_CALL( voronoi_zone_mock3, get_id() ).WillByDefault(Return(0));
    EXPECT_CALL( voronoi_zone_mock3, get_attribute_range() ).WillOnce(Return(list_of(3.0)(30.0)));

    NiceMock<voronoi_zone_mock> voronoi_zone_mock4;
	ON_CALL( voronoi_zone_mock4, get_id() ).WillByDefault(Return(0));
    EXPECT_CALL( voronoi_zone_mock4, get_attribute_range() ).WillOnce(Return(list_of(4.0)(40.0)));

	// initialize zone
	zone_type zone(cref_list_of<4>(voronoi_zone_mock1)(voronoi_zone_mock2)(voronoi_zone_mock3)(voronoi_zone_mock4));

	ASSERT_THAT( zone.get_means(), EqDoubleRange(list_of(2.5)(25.0)) );
	ASSERT_THAT( zone.get_variances(), EqDoubleRange(list_of(1.25)(125.0)) );
	ASSERT_THAT( zone.get_standard_deviations(), EqDoubleRange(transform(list_of(1.25)(125.0), sqrt_computer())) );
	ASSERT_THROW( zone.get_mean(2), std::out_of_range );
	ASSERT_THROW( zone.get_variance(2), std::out_of_range );
	ASSERT_THROW( zone.get_standard_deviation(2), std::out_of_range );
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

TEST(zone_tester, test_merge) {

    // initialize voronoi zones
	voronoi_zone_mock voronoi_zone_mock1;
	polygon_type geometry1 = make_polygon<kernel_type>(list_of(point_type(0, 0))(point_type(1, 0))(point_type(1, 1))(point_type(0, 1)));
	EXPECT_CALL( voronoi_zone_mock1, get_id() ).WillRepeatedly(Return(3));
    EXPECT_CALL( voronoi_zone_mock1, get_geometry() ).Times(2).WillRepeatedly(Return(geometry1));
    EXPECT_CALL( voronoi_zone_mock1, get_attribute_range() ).Times(2).WillRepeatedly(Return(list_of(0.0)(1.0)));

	voronoi_zone_mock voronoi_zone_mock2;
	polygon_type geometry2 = make_polygon<kernel_type>(list_of(point_type(1, 0))(point_type(2, 0))(point_type(2, 1))(point_type(1, 1)));
	EXPECT_CALL( voronoi_zone_mock2, get_id() ).WillRepeatedly(Return(4));
    EXPECT_CALL( voronoi_zone_mock2, get_geometry() ).Times(2).WillRepeatedly(Return(geometry2));
    EXPECT_CALL( voronoi_zone_mock2, get_attribute_range() ).Times(2).WillRepeatedly(Return(list_of(2.0)(3.0)));

	voronoi_zone_mock voronoi_zone_mock3;
	polygon_type geometry3 = make_polygon<kernel_type>(list_of(point_type(0, 1))(point_type(1, 1))(point_type(1, 2))(point_type(0, 2)));
	EXPECT_CALL( voronoi_zone_mock3, get_id() ).WillRepeatedly(Return(1));
    EXPECT_CALL( voronoi_zone_mock3, get_geometry() ).WillOnce(Return(geometry3));
    EXPECT_CALL( voronoi_zone_mock3, get_attribute_range() ).WillOnce(Return(list_of(4.0)(5.0)));

	voronoi_zone_mock voronoi_zone_mock4;
	polygon_type geometry4 = make_polygon<kernel_type>(list_of(point_type(1, 1))(point_type(2, 1))(point_type(2, 2))(point_type(1, 2)));
	EXPECT_CALL( voronoi_zone_mock4, get_id() ).WillRepeatedly(Return(2));
    EXPECT_CALL( voronoi_zone_mock4, get_geometry() ).WillOnce(Return(geometry4));
    EXPECT_CALL( voronoi_zone_mock4, get_attribute_range() ).WillOnce(Return(list_of(6.0)(7.0)));

    // initialize mergeable zone
	zone_type zone_mergeable(cref_list_of<2>(voronoi_zone_mock1)(voronoi_zone_mock2));
	ASSERT_EQ( 3, zone_mergeable.get_id() );
	ASSERT_EQ( make_polygon_with_holes<kernel_type>(list_of(point_type(0, 0))(point_type(1, 0))(point_type(2, 0))(point_type(2, 1))(point_type(1, 1))(point_type(0, 1))), zone_mergeable.get_geometry() );
	ASSERT_THAT( zone_mergeable.get_means(), EqDoubleRange(list_of(1.0)(2.0)) );

    // initialize merged zone
	zone_type zone_merged(cref_list_of<2>(voronoi_zone_mock3)(voronoi_zone_mock4));
	ASSERT_EQ( 1, zone_merged.get_id() );

	// merge zone
	zone_mergeable.merge(zone_merged);
	ASSERT_EQ( (size_t)4, zone_mergeable.size() );
	ASSERT_EQ( 1, zone_mergeable.get_id() );
	ASSERT_EQ( make_polygon_with_holes<kernel_type>(list_of(point_type(0, 0))(point_type(1, 0))(point_type(2, 0))(point_type(2, 1))(point_type(2, 2))(point_type(1, 2))(point_type(0, 2))(point_type(0, 1))), zone_mergeable.get_geometry() );
	ASSERT_THAT( zone_mergeable.get_means(), EqDoubleRange(list_of(3.0)(4.0)) );
}
