/**
 * GeoFIS Native Library
 *
 * Copyright (C) 2013 IRSTEA INRA SupAgro
 *
 * IDDN.FR.001.390016.000.R.P.2013.000.31235
 *
 * Auteurs/Authors:
 *     Jean-luc Lablée - IRSTEA
 *     Serge Guillaume - IRSTEA
 *     Brigitte Charnomordic - INRA
 *     Anne Tireau - INRA
 *     Bruno Tisseyre - SupAgro
 *
 * =========================================================================
 *
 * Ce logiciel est régi par la licence CeCILL soumise au droit français et
 * respectant les principes de diffusion des logiciels libres. Vous pouvez
 * utiliser, modifier et/ou redistribuer ce programme sous les conditions
 * de la licence CeCILL telle que diffusée par le CEA, le CNRS et l'INRIA
 * sur le site "http://www.cecill.info".
 *
 * En contrepartie de l'accessibilité au code source et des droits de copie,
 * de modification et de redistribution accordés par cette licence, il n'est
 * offert aux utilisateurs qu'une garantie limitée.  Pour les mêmes raisons,
 * seule une responsabilité restreinte pèse sur l'auteur du programme,  le
 * titulaire des droits patrimoniaux et les concédants successifs.
 *
 * A cet égard  l'attention de l'utilisateur est attirée sur les risques
 * associés au chargement,  à l'utilisation,  à la modification et/ou au
 * développement et à la reproduction du logiciel par l'utilisateur étant
 * donné sa spécificité de logiciel libre, qui peut le rendre complexe à
 * manipuler et qui le réserve donc à des développeurs et des professionnels
 * avertis possédant  des  connaissances  informatiques approfondies.  Les
 * utilisateurs sont donc invités à charger  et  tester  l'adéquation  du
 * logiciel à leurs besoins dans des conditions permettant d'assurer la
 * sécurité de leurs systèmes et ou de leurs données et, plus généralement,
 * à l'utiliser et l'exploiter dans les mêmes conditions de sécurité.
 *
 * Le fait que vous puissiez accéder à cet en-tête signifie que vous avez
 * pris connaissance de la licence CeCILL, et que vous en avez accepté les
 * termes.
 *
 * =========================================================================
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
#include <geofis/algorithm/zoning/map/map.hpp>
#include <boost/assign/list_of.hpp>
#include <boost/utility/addressof.hpp>
#include <boost/noncopyable.hpp>
#include <util/gmock/matchers.hpp>
#include <gmock/gmock.h>

using namespace geofis;
using namespace boost;
using namespace boost::assign;
using namespace testing;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

namespace {

struct zone;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

struct zone_mock  {

	MOCK_METHOD1( copy, void (const zone &) );
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

struct zone {

	zone(int id, zone_mock &mock) : id(id), mock(mock) {}
	zone(const zone &other) : id(other.id), mock(other.mock) {
		mock.copy(other);
	}

	friend bool operator==(const zone &lhs, const zone &rhs) {
		return lhs.id == rhs.id;
	}

	friend bool operator!=(const zone &lhs, const zone &rhs) {
		return !(lhs == rhs);
	}

	int id;
	zone_mock &mock;
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

struct zone_non_copyable : public noncopyable {

	friend bool operator==(const zone_non_copyable &lhs, const zone_non_copyable &rhs) {
		return addressof(lhs) == addressof(rhs);
	}

	friend bool operator!=(const zone_non_copyable &lhs, const zone_non_copyable &rhs) {
		return !(lhs == rhs);
	}
};

} // namespace

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

typedef map<zone> map_type;
typedef map<reference_wrapper<zone_non_copyable> > map_reference_wrapper_type;
typedef map<reference_wrapper<const zone_non_copyable> > map_const_reference_wrapper_type;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

TEST(map_tester, test_map) {

	zone_mock zone_mock;
	EXPECT_CALL( zone_mock, copy(_) ).Times(AnyNumber());
	zone zone1(1, zone_mock), zone2(2, zone_mock), zone3(3, zone_mock), zone4(4, zone_mock);
	map_type map(list_of(zone1)(zone2)(zone3)(zone4));

	EXPECT_CALL( zone_mock, copy(_) ).Times(0); // no more copies in map

	ASSERT_EQ( (size_t)4, map.size() );
	ASSERT_THAT( map.get_zones(), EqRange(ref_list_of<4>(zone1)(zone2)(zone3)(zone4)) );
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

TEST(map_tester, test_reference_wraper_map) {

	zone_non_copyable zone1, zone2, zone3, zone4;
	map_reference_wrapper_type map(ref_list_of<4>(zone1)(zone2)(zone3)(zone4));

	ASSERT_EQ( (size_t)4, map.size() );
	ASSERT_THAT( map.get_zones(), EqRange(ref_list_of<4>(zone1)(zone2)(zone3)(zone4)) );
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

TEST(map_tester, test_const_reference_wraper_map) {

	zone_non_copyable zone1, zone2, zone3, zone4;
	map_const_reference_wrapper_type map(ref_list_of<4>(zone1)(zone2)(zone3)(zone4));

	ASSERT_EQ( (size_t)4, map.size() );
	ASSERT_THAT( map.get_zones(), EqRange(cref_list_of<4>(zone1)(zone2)(zone3)(zone4)) );
}
