/**
 * GeoFIS Native Library
 *
 * Copyright (C) 2013 IRSTEA INRA SupAgro
 *
 * IDDN.FR.001.390016.000.R.P.2013.000.31235
 *
 * Auteurs/Authors:
 *     Jean-luc Lablée - IRSTEA
 *     Serge Guillaume - IRSTEA
 *     Brigitte Charnomordic - INRA
 *     Anne Tireau - INRA
 *     Bruno Tisseyre - SupAgro
 *
 * =========================================================================
 *
 * Ce logiciel est régi par la licence CeCILL soumise au droit français et
 * respectant les principes de diffusion des logiciels libres. Vous pouvez
 * utiliser, modifier et/ou redistribuer ce programme sous les conditions
 * de la licence CeCILL telle que diffusée par le CEA, le CNRS et l'INRIA
 * sur le site "http://www.cecill.info".
 *
 * En contrepartie de l'accessibilité au code source et des droits de copie,
 * de modification et de redistribution accordés par cette licence, il n'est
 * offert aux utilisateurs qu'une garantie limitée.  Pour les mêmes raisons,
 * seule une responsabilité restreinte pèse sur l'auteur du programme,  le
 * titulaire des droits patrimoniaux et les concédants successifs.
 *
 * A cet égard  l'attention de l'utilisateur est attirée sur les risques
 * associés au chargement,  à l'utilisation,  à la modification et/ou au
 * développement et à la reproduction du logiciel par l'utilisateur étant
 * donné sa spécificité de logiciel libre, qui peut le rendre complexe à
 * manipuler et qui le réserve donc à des développeurs et des professionnels
 * avertis possédant  des  connaissances  informatiques approfondies.  Les
 * utilisateurs sont donc invités à charger  et  tester  l'adéquation  du
 * logiciel à leurs besoins dans des conditions permettant d'assurer la
 * sécurité de leurs systèmes et ou de leurs données et, plus généralement,
 * à l'utiliser et l'exploiter dans les mêmes conditions de sécurité.
 *
 * Le fait que vous puissiez accéder à cet en-tête signifie que vous avez
 * pris connaissance de la licence CeCILL, et que vous en avez accepté les
 * termes.
 *
 * =========================================================================
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
#include <geofis/algorithm/zoning/merging/merging.hpp>
#include <util/functional/binary_reference_adaptor.hpp>
#include <util/functional/unary_reference_adaptor.hpp>
#include <util/range/algorithm/copy_iterator_if.hpp>
#include <util/gmock/matchers.hpp>
#include <boost/range/algorithm/min_element.hpp>
#include <boost/assign/list_of.hpp>
#include <gmock/gmock.h>

using namespace util;
using namespace geofis;
using namespace boost;
using namespace boost::assign;
using namespace testing;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

namespace {

struct zone;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

struct zone_mock  {

	MOCK_METHOD2( merge, void (const zone &mergeable_zone, const zone &merged_zone) );
	MOCK_METHOD1( copy, void (const zone &) );
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

struct zone {

	zone(int id, zone_mock &mock) : id(id), mock(mock) {}
	zone(const zone &other) : id(other.id), mock(other.mock) {
		mock.copy(other);
	}

	friend bool operator==(const zone &lhs, const zone &rhs) {
		return lhs.id == rhs.id;
	}

	friend bool operator!=(const zone &lhs, const zone &rhs) {
		return !(lhs == rhs);
	}

	void merge(const zone &merged_zone) {
		mock.merge(*this, merged_zone);
	}

	int id;
	zone_mock &mock;
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

struct mergeable_predicate_mock {

	typedef bool result_type;

	MOCK_CONST_METHOD1( unary_operator, result_type (const zone &) );

	result_type operator()(const zone &zone) const {
		return unary_operator(zone);
	}
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

struct zone_neighbor_predicate_mock {

	typedef bool result_type;

	MOCK_CONST_METHOD2( binary_operator, result_type (const zone &mergeable_zone, const zone &merged_zone) );

	template <class Zone> result_type operator()(const Zone &mergeable_zone, const Zone &merged_zone) const {
		return binary_operator(mergeable_zone, merged_zone);
	}
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

struct zone_distance_mock {

	typedef double result_type;

	MOCK_CONST_METHOD2( binary_operator, result_type (const zone &mergeable_zone, const zone &merged_zone) );

	template <class Zone> result_type operator()(const Zone &mergeable_zone, const Zone &merged_zone) const {
		return binary_operator(mergeable_zone, merged_zone);
	}
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

struct zone_pair_merger_mock {

	MOCK_CONST_METHOD1( merge_distance, bool (double distance) );

	template <class ZonePair> bool operator()(const ZonePair &zone_pair) const {
		return merge_distance(zone_pair.get_distance());
	}

	template <class ZonePairRange, class OutputIterator> void operator()(ZonePairRange &zone_pairs, OutputIterator output) const {
		util::copy_iterator_if(zone_pairs, output, *this);
	}
};

} // namespace

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

typedef merging<zone> merging_type;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

TEST(merging_tester, test_mergeable_zones_empty) {

	zone_mock zone_mock;
	EXPECT_CALL( zone_mock, copy(_) ).Times(2); // zone copies in merging

	zone zone1(1, zone_mock), zone2(2, zone_mock);

	mergeable_predicate_mock mergeable_predicate_mock;
	EXPECT_CALL( mergeable_predicate_mock, unary_operator(_) ).WillRepeatedly(Return(false));

	merging_type merging(make_unary_reference_adaptor(mergeable_predicate_mock), cref_list_of<2>(zone1)(zone2));

	ASSERT_FALSE( merging.is_merged_zones_empty() );
	ASSERT_EQ( (size_t)2, merging.get_merged_zones_size() );

	ASSERT_TRUE( merging.is_mergeable_zones_empty() );
	ASSERT_EQ( (size_t)0, merging.get_mergeable_zones_size() );
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

TEST(merging_tester, test_merged_zones_empty) {

	zone_mock zone_mock;
	EXPECT_CALL( zone_mock, copy(_) ).Times(2);

	zone zone1(1, zone_mock), zone2(2, zone_mock);

	mergeable_predicate_mock mergeable_predicate_mock;
	EXPECT_CALL( mergeable_predicate_mock, unary_operator(_) ).WillRepeatedly(Return(true));

	merging_type merging(make_unary_reference_adaptor(mergeable_predicate_mock), cref_list_of<2>(zone1)(zone2));

	ASSERT_TRUE( merging.is_merged_zones_empty() );
	ASSERT_EQ( (size_t)0, merging.get_merged_zones_size() );

	ASSERT_FALSE( merging.is_mergeable_zones_empty() );
	ASSERT_EQ( (size_t)2, merging.get_mergeable_zones_size() );
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

template <class MergedZoneRange, class MergeableZoneRange> void assert_merge_partition(const merging_type &merging, const MergedZoneRange &expected_merged_zones, const MergeableZoneRange &expected_mergeable_zones) {

	ASSERT_EQ( boost::size(expected_merged_zones), merging.get_merged_zones_size() );
	ASSERT_EQ( boost::size(expected_mergeable_zones), merging.get_mergeable_zones_size() );
	ASSERT_THAT( merging.get_merged_zones(), EqRange(expected_merged_zones) );
	ASSERT_THAT( merging.get_mergeable_zones(), EqRange(expected_mergeable_zones) );
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

TEST(merging_tester, test_partition) {

	zone_mock zone_mock;
	EXPECT_CALL( zone_mock, copy(_) ).Times(5);

	zone zone1(1, zone_mock), zone2(2, zone_mock), zone3(3, zone_mock), zone4(4, zone_mock), zone5(5, zone_mock);

	mergeable_predicate_mock mergeable_predicate_mock;
	EXPECT_CALL( mergeable_predicate_mock, unary_operator(_) ).WillRepeatedly(Return(true));
	EXPECT_CALL( mergeable_predicate_mock, unary_operator(Eq(ByRef(zone2))) ).WillOnce(Return(false));
	EXPECT_CALL( mergeable_predicate_mock, unary_operator(Eq(ByRef(zone4))) ).WillOnce(Return(false));

	merging_type merging(make_unary_reference_adaptor(mergeable_predicate_mock), cref_list_of<5>(zone1)(zone2)(zone3)(zone4)(zone5));
	assert_merge_partition( merging, cref_list_of<2>(zone2)(zone4), cref_list_of<3>(zone1)(zone3)(zone5));
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/**
 *   ┌───────┬───────┐
 *   │   1   │   2   │
 *   │ ┌───┐ │ ┌───┐ │
 *   │ │ 3 │ │ │ 4 │ │
 *   │ └───┘ │ └───┘ │
 *   └───────┴───────┘
 *   mergeable zones: 1, 2
 *   merged zones: 3, 4
 */
TEST(merging_tester, test_merge_configuration_1) {

	zone_mock zone_mock;
	EXPECT_CALL( zone_mock, copy(_) ).Times(4);

	zone zone1(1, zone_mock), zone2(2, zone_mock), zone3(3, zone_mock), zone4(4, zone_mock);

	mergeable_predicate_mock mergeable_predicate_mock;
	EXPECT_CALL( mergeable_predicate_mock, unary_operator(_) ).WillRepeatedly(Return(true));
	EXPECT_CALL( mergeable_predicate_mock, unary_operator(Eq(ByRef(zone3))) ).WillOnce(Return(false));
	EXPECT_CALL( mergeable_predicate_mock, unary_operator(Eq(ByRef(zone4))) ).WillOnce(Return(false));

	merging_type merging(make_unary_reference_adaptor(mergeable_predicate_mock), cref_list_of<4>(zone1)(zone2)(zone3)(zone4));
	assert_merge_partition( merging, cref_list_of<2>(zone3)(zone4), cref_list_of<2>(zone1)(zone2));
	EXPECT_CALL( zone_mock, copy(_) ).Times(0); // no more zone copies in merging

	zone_neighbor_predicate_mock zone_neighbor_predicate_mock;
	EXPECT_CALL( zone_neighbor_predicate_mock, binary_operator(_, _) ).WillRepeatedly(Return(false));
	EXPECT_CALL( zone_neighbor_predicate_mock, binary_operator(Eq(ByRef(zone1)), Eq(ByRef(zone3))) ).WillOnce(Return(true));
	EXPECT_CALL( zone_neighbor_predicate_mock, binary_operator(Eq(ByRef(zone2)), Eq(ByRef(zone4))) ).WillOnce(Return(true));

	InSequence merge_order;
	EXPECT_CALL( zone_mock, merge(Eq(ByRef(zone1)), Eq(ByRef(zone3))) );
	EXPECT_CALL( zone_mock, merge(Eq(ByRef(zone2)), Eq(ByRef(zone4))) );

	merging.compute_merge_zones(make_binary_reference_adaptor(zone_neighbor_predicate_mock), zone_distance_mock(), zone_pair_merger_mock());
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/**
 * ┌───────────┐
 * │     1     │
 * │ ┌───────┐ │
 * │ │   2   │ │
 * │ │ ┌───┐ │ │
 * │ │ │ 3 │ │ │
 * │ │ └───┘ │ │
 * │ └───────┘ │
 * └───────────┘
 *   mergeable zones: 1
 *   merged zones: 2, 3
 */
TEST(merging_tester, test_merge_configuration_2) {

	zone_mock zone_mock;
	EXPECT_CALL( zone_mock, copy(_) ).Times(3);

	zone zone1(1, zone_mock), zone2(2, zone_mock), zone3(3, zone_mock);

	mergeable_predicate_mock mergeable_predicate_mock;
	EXPECT_CALL( mergeable_predicate_mock, unary_operator(_) ).WillRepeatedly(Return(true));
	EXPECT_CALL( mergeable_predicate_mock, unary_operator(Eq(ByRef(zone2))) ).WillOnce(Return(false));
	EXPECT_CALL( mergeable_predicate_mock, unary_operator(Eq(ByRef(zone3))) ).WillOnce(Return(false));

	merging_type merging(make_unary_reference_adaptor(mergeable_predicate_mock), cref_list_of<3>(zone1)(zone2)(zone3));
	assert_merge_partition( merging, cref_list_of<2>(zone2)(zone3), cref_list_of<1>(zone1));
	EXPECT_CALL( zone_mock, copy(_) ).Times(0); // no more zone copies in merging

	zone_neighbor_predicate_mock zone_neighbor_predicate_mock;
	EXPECT_CALL( zone_neighbor_predicate_mock, binary_operator(_, _) ).WillRepeatedly(Return(false));
	EXPECT_CALL( zone_neighbor_predicate_mock, binary_operator(Eq(ByRef(zone1)), Eq(ByRef(zone2))) ).WillOnce(Return(true));
	EXPECT_CALL( zone_neighbor_predicate_mock, binary_operator(Eq(ByRef(zone1)), Eq(ByRef(zone3))) ).WillOnce(Return(false)).WillOnce(Return(true));

	InSequence merge_order;
	EXPECT_CALL( zone_mock, merge(Eq(ByRef(zone1)), Eq(ByRef(zone2))) );
	EXPECT_CALL( zone_mock, merge(Eq(ByRef(zone1)), Eq(ByRef(zone3))) );

	merging.compute_merge_zones(make_binary_reference_adaptor(zone_neighbor_predicate_mock), zone_distance_mock(), zone_pair_merger_mock());
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/**
 * ┌───────────┐
 * │     1     │
 * │ ┌───────┐ │
 * │ │   3   │ │
 * │ │ ┌───┐ │ │
 * │ │ │ 2 │ │ │
 * │ │ └───┘ │ │
 * │ └───────┘ │
 * └───────────┘
 *   mergeable zones: 1
 *   merged zones: 2, 3
 */
TEST(merging_tester, test_merge_configuration_3) {

	zone_mock zone_mock;
	EXPECT_CALL( zone_mock, copy(_) ).Times(3);

	zone zone1(1, zone_mock), zone2(2, zone_mock), zone3(3, zone_mock);

	mergeable_predicate_mock mergeable_predicate_mock;
	EXPECT_CALL( mergeable_predicate_mock, unary_operator(_) ).WillRepeatedly(Return(true));
	EXPECT_CALL( mergeable_predicate_mock, unary_operator(Eq(ByRef(zone2))) ).WillOnce(Return(false));
	EXPECT_CALL( mergeable_predicate_mock, unary_operator(Eq(ByRef(zone3))) ).WillOnce(Return(false));

	merging_type merging(make_unary_reference_adaptor(mergeable_predicate_mock), cref_list_of<3>(zone1)(zone2)(zone3));
	assert_merge_partition( merging, cref_list_of<2>(zone2)(zone3), cref_list_of<1>(zone1));
	EXPECT_CALL( zone_mock, copy(_) ).Times(0); // no more zone copies in merging

	zone_neighbor_predicate_mock zone_neighbor_predicate_mock;
	EXPECT_CALL( zone_neighbor_predicate_mock, binary_operator(_, _) ).WillRepeatedly(Return(false));
	EXPECT_CALL( zone_neighbor_predicate_mock, binary_operator(Eq(ByRef(zone1)), Eq(ByRef(zone2))) ).WillOnce(Return(false)).WillOnce(Return(true));
	EXPECT_CALL( zone_neighbor_predicate_mock, binary_operator(Eq(ByRef(zone1)), Eq(ByRef(zone3))) ).WillOnce(Return(true));

	InSequence merge_order;
	EXPECT_CALL( zone_mock, merge(Eq(ByRef(zone1)), Eq(ByRef(zone3))) );
	EXPECT_CALL( zone_mock, merge(Eq(ByRef(zone1)), Eq(ByRef(zone2))) );

	merging.compute_merge_zones(make_binary_reference_adaptor(zone_neighbor_predicate_mock), zone_distance_mock(), zone_pair_merger_mock());
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/**
 *   ┌─────┬────┐
 *   │   1 │  2 │
 *   │ ┌───┤    │
 *   │ │ 3 │    │
 *   │ └───┤    │
 *   └─────┴────┘
 *   mergeable zones: 1, 2
 *   merged zones: 3
 */
TEST(merging_tester, test_merge_configuration_4) {

	zone_mock zone_mock;
	EXPECT_CALL( zone_mock, copy(_) ).Times(3);

	zone zone1(1, zone_mock), zone2(2, zone_mock), zone3(3, zone_mock);

	mergeable_predicate_mock mergeable_predicate_mock;
	EXPECT_CALL( mergeable_predicate_mock, unary_operator(_) ).WillRepeatedly(Return(true));
	EXPECT_CALL( mergeable_predicate_mock, unary_operator(Eq(ByRef(zone3))) ).WillOnce(Return(false));

	merging_type merging(make_unary_reference_adaptor(mergeable_predicate_mock), cref_list_of<3>(zone1)(zone2)(zone3));
	assert_merge_partition( merging, cref_list_of<1>(zone3), cref_list_of<2>(zone1)(zone2));
	EXPECT_CALL( zone_mock, copy(_) ).Times(0); // no more zone copies in merging

	zone_neighbor_predicate_mock zone_neighbor_predicate_mock;
	EXPECT_CALL( zone_neighbor_predicate_mock, binary_operator(_, _) ).WillRepeatedly(Return(false));
	EXPECT_CALL( zone_neighbor_predicate_mock, binary_operator(Eq(ByRef(zone1)), Eq(ByRef(zone3))) ).WillOnce(Return(true));
	EXPECT_CALL( zone_neighbor_predicate_mock, binary_operator(Eq(ByRef(zone2)), Eq(ByRef(zone3))) ).WillOnce(Return(true));

	zone_distance_mock zone_distance_mock;
	EXPECT_CALL( zone_distance_mock, binary_operator(Eq(ByRef(zone1)), Eq(ByRef(zone3))) ).WillOnce(Return(1));
	EXPECT_CALL( zone_distance_mock, binary_operator(Eq(ByRef(zone2)), Eq(ByRef(zone3))) ).WillOnce(Return(2));

	zone_pair_merger_mock zone_pair_merger_mock;
	EXPECT_CALL( zone_pair_merger_mock, merge_distance(_) ).WillRepeatedly(Return(false));
	EXPECT_CALL( zone_pair_merger_mock, merge_distance(Eq(1)) ).WillRepeatedly(Return(true));

	EXPECT_CALL( zone_mock, merge(Eq(ByRef(zone1)), Eq(ByRef(zone3))) );

	merging.compute_merge_zones(make_binary_reference_adaptor(zone_neighbor_predicate_mock), zone_distance_mock, zone_pair_merger_mock);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/**
 *   ┌─────┬────┐
 *   │   1 │  2 │
 *   │ ┌───┤    │
 *   │ │ 4 ├────┤
 *   │ └───┤  3 │
 *   └─────┴────┘
 *   mergeable zones: 1, 2, 3
 *   merged zones: 4
 */
TEST(merging_tester, test_merge_configuration_5) {

	zone_mock zone_mock;
	EXPECT_CALL( zone_mock, copy(_) ).Times(4);

	zone zone1(1, zone_mock), zone2(2, zone_mock), zone3(3, zone_mock), zone4(4, zone_mock);

	mergeable_predicate_mock mergeable_predicate_mock;
	EXPECT_CALL( mergeable_predicate_mock, unary_operator(_) ).WillRepeatedly(Return(true));
	EXPECT_CALL( mergeable_predicate_mock, unary_operator(Eq(ByRef(zone4))) ).WillOnce(Return(false));

	merging_type merging(make_unary_reference_adaptor(mergeable_predicate_mock), cref_list_of<4>(zone1)(zone2)(zone3)(zone4));
	assert_merge_partition( merging, cref_list_of<1>(zone4), cref_list_of<3>(zone1)(zone2)(zone3));
	EXPECT_CALL( zone_mock, copy(_) ).Times(0); // no more zone copies in merging

	zone_neighbor_predicate_mock zone_neighbor_predicate_mock;
	EXPECT_CALL( zone_neighbor_predicate_mock, binary_operator(_, _) ).WillRepeatedly(Return(false));
	EXPECT_CALL( zone_neighbor_predicate_mock, binary_operator(Eq(ByRef(zone1)), Eq(ByRef(zone4))) ).WillOnce(Return(true));
	EXPECT_CALL( zone_neighbor_predicate_mock, binary_operator(Eq(ByRef(zone2)), Eq(ByRef(zone4))) ).WillOnce(Return(true));
	EXPECT_CALL( zone_neighbor_predicate_mock, binary_operator(Eq(ByRef(zone3)), Eq(ByRef(zone4))) ).WillOnce(Return(true));

	zone_distance_mock zone_distance_mock;
	EXPECT_CALL( zone_distance_mock, binary_operator(Eq(ByRef(zone1)), Eq(ByRef(zone4))) ).WillOnce(Return(2));
	EXPECT_CALL( zone_distance_mock, binary_operator(Eq(ByRef(zone2)), Eq(ByRef(zone4))) ).WillOnce(Return(1));
	EXPECT_CALL( zone_distance_mock, binary_operator(Eq(ByRef(zone3)), Eq(ByRef(zone4))) ).WillOnce(Return(1));

	zone_pair_merger_mock zone_pair_merger_mock;
	EXPECT_CALL( zone_pair_merger_mock, merge_distance(_) ).WillRepeatedly(Return(false));
	EXPECT_CALL( zone_pair_merger_mock, merge_distance(Eq(1)) ).WillRepeatedly(Return(true));

	EXPECT_CALL( zone_mock, merge(Eq(ByRef(zone2)), Eq(ByRef(zone4))) );

	merging.compute_merge_zones(make_binary_reference_adaptor(zone_neighbor_predicate_mock), zone_distance_mock, zone_pair_merger_mock);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/**
 *   ┌─────┬─────┐
 *   │   1 │  2  │
 *   │ ┌───┤     │
 *   │ │ 5 ├─────┤
 *   │ └───┤  3  │
 *   │     │┌───┐│
 *   │     ││ 4 ││
 *   │     │└───┘│
 *   └─────┴─────┘
 *   mergeable zones: 1, 2, 3
 *   merged zones: 4, 5
 */
TEST(merging_tester, test_merge_configuration_6) {

	zone_mock zone_mock;
	EXPECT_CALL( zone_mock, copy(_) ).Times(5);

	zone zone1(1, zone_mock), zone2(2, zone_mock), zone3(3, zone_mock), zone4(4, zone_mock), zone5(5, zone_mock);

	mergeable_predicate_mock mergeable_predicate_mock;
	EXPECT_CALL( mergeable_predicate_mock, unary_operator(_) ).WillRepeatedly(Return(true));
	EXPECT_CALL( mergeable_predicate_mock, unary_operator(Eq(ByRef(zone4))) ).WillOnce(Return(false));
	EXPECT_CALL( mergeable_predicate_mock, unary_operator(Eq(ByRef(zone5))) ).WillOnce(Return(false));

	merging_type merging(make_unary_reference_adaptor(mergeable_predicate_mock), cref_list_of<5>(zone1)(zone2)(zone3)(zone4)(zone5));
	assert_merge_partition( merging, cref_list_of<2>(zone4)(zone5), cref_list_of<3>(zone1)(zone2)(zone3));
	EXPECT_CALL( zone_mock, copy(_) ).Times(0); // no more zone copies in merging

	zone_neighbor_predicate_mock zone_neighbor_predicate_mock;
	EXPECT_CALL( zone_neighbor_predicate_mock, binary_operator(_, _) ).WillRepeatedly(Return(false));
	EXPECT_CALL( zone_neighbor_predicate_mock, binary_operator(Eq(ByRef(zone3)), Eq(ByRef(zone4))) ).WillOnce(Return(true));
	EXPECT_CALL( zone_neighbor_predicate_mock, binary_operator(Eq(ByRef(zone1)), Eq(ByRef(zone5))) ).WillOnce(Return(true));
	EXPECT_CALL( zone_neighbor_predicate_mock, binary_operator(Eq(ByRef(zone2)), Eq(ByRef(zone5))) ).WillOnce(Return(true));
	EXPECT_CALL( zone_neighbor_predicate_mock, binary_operator(Eq(ByRef(zone3)), Eq(ByRef(zone5))) ).WillOnce(Return(true));

	zone_distance_mock zone_distance_mock;
	EXPECT_CALL( zone_distance_mock, binary_operator(Eq(ByRef(zone1)), Eq(ByRef(zone5))) ).WillOnce(Return(2));
	EXPECT_CALL( zone_distance_mock, binary_operator(Eq(ByRef(zone2)), Eq(ByRef(zone5))) ).WillOnce(Return(1));
	EXPECT_CALL( zone_distance_mock, binary_operator(Eq(ByRef(zone3)), Eq(ByRef(zone5))) ).WillOnce(Return(1));

	zone_pair_merger_mock zone_pair_merger_mock;
	EXPECT_CALL( zone_pair_merger_mock, merge_distance(_) ).WillRepeatedly(Return(false));
	EXPECT_CALL( zone_pair_merger_mock, merge_distance(Eq(1)) ).WillRepeatedly(Return(true));

	InSequence merge_order;
	EXPECT_CALL( zone_mock, merge(Eq(ByRef(zone3)), Eq(ByRef(zone4))) );
	EXPECT_CALL( zone_mock, merge(Eq(ByRef(zone2)), Eq(ByRef(zone5))) );

	merging.compute_merge_zones(make_binary_reference_adaptor(zone_neighbor_predicate_mock), zone_distance_mock, zone_pair_merger_mock);
}
