/**
 * GeoFIS Native Library
 *
 * Copyright (C) 2013 IRSTEA INRA SupAgro
 *
 * IDDN.FR.001.390016.000.R.P.2013.000.31235
 *
 * Auteurs/Authors:
 *     Jean-luc Lablée - IRSTEA
 *     Serge Guillaume - IRSTEA
 *     Brigitte Charnomordic - INRA
 *     Anne Tireau - INRA
 *     Bruno Tisseyre - SupAgro
 *
 * =========================================================================
 *
 * Ce logiciel est régi par la licence CeCILL soumise au droit français et
 * respectant les principes de diffusion des logiciels libres. Vous pouvez
 * utiliser, modifier et/ou redistribuer ce programme sous les conditions
 * de la licence CeCILL telle que diffusée par le CEA, le CNRS et l'INRIA
 * sur le site "http://www.cecill.info".
 *
 * En contrepartie de l'accessibilité au code source et des droits de copie,
 * de modification et de redistribution accordés par cette licence, il n'est
 * offert aux utilisateurs qu'une garantie limitée.  Pour les mêmes raisons,
 * seule une responsabilité restreinte pèse sur l'auteur du programme,  le
 * titulaire des droits patrimoniaux et les concédants successifs.
 *
 * A cet égard  l'attention de l'utilisateur est attirée sur les risques
 * associés au chargement,  à l'utilisation,  à la modification et/ou au
 * développement et à la reproduction du logiciel par l'utilisateur étant
 * donné sa spécificité de logiciel libre, qui peut le rendre complexe à
 * manipuler et qui le réserve donc à des développeurs et des professionnels
 * avertis possédant  des  connaissances  informatiques approfondies.  Les
 * utilisateurs sont donc invités à charger  et  tester  l'adéquation  du
 * logiciel à leurs besoins dans des conditions permettant d'assurer la
 * sécurité de leurs systèmes et ou de leurs données et, plus généralement,
 * à l'utiliser et l'exploiter dans les mêmes conditions de sécurité.
 *
 * Le fait que vous puissiez accéder à cet en-tête signifie que vous avez
 * pris connaissance de la licence CeCILL, et que vous en avez accepté les
 * termes.
 *
 * =========================================================================
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
#include <geofis/ogr/ogr_directory_data_source.hpp>
#include <geofis/ogr/ogr_data_source.hpp>
#include <geofis/ogr/ogr_layer.hpp>
#include <geofis/ogr/ogr_feature.hpp>
#include <geofis/ogr/ogr_spatial_reference.hpp>
#include <geofis/ogr/shapefile/ogr_shapefile.hpp>
#include <geofis/geometry/polygon.hpp>
#include <geofis/geometry/polygon_with_holes.hpp>
#include <boost/assign/list_of.hpp>
#include <boost/filesystem/remove_all.hpp>
#include <CGAL/Exact_predicates_exact_constructions_kernel.h>
#include <cppunit/extensions/HelperMacros.h>
#include <cppunit/specializations.hpp>
#include <cppunit/Exception.h>

#include <iostream>

using namespace geofis;
using namespace boost;
using namespace boost::filesystem;
using namespace boost::assign;
using namespace CGAL;

typedef Exact_predicates_exact_constructions_kernel kernel_type;
typedef Point_2<kernel_type> point_type;
typedef Polygon_2<kernel_type> polygon_type;
typedef Polygon_with_holes_2<kernel_type> polygon_with_holes_type;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class ogr_tester : public CppUnit::TestFixture {

	CPPUNIT_TEST_SUITE( ogr_tester );

	CPPUNIT_TEST( test_ogr_directory_data_source );
	CPPUNIT_TEST( test_ogr_point );
	CPPUNIT_TEST( test_ogr_polygon );
	CPPUNIT_TEST( test_ogr_polygon_with_holes );

	CPPUNIT_TEST_SUITE_END();

#if GDAL_VERSION_MAJOR >= 2
	GDALDriver *driver;
#else
	OGRSFDriver *driver;
#endif
	OGRSpatialReference spatial_reference;

public:

	void setUp() {
		driver = get_ogr_shapefile_driver();
		spatial_reference.importFromEPSG(27563);
	}

	void test_ogr_directory_data_source() {
		try {
			ogr_directory_data_source data_source("test", driver, TRUE);
			data_source.delete_layers();
			ogr_layer layer1(data_source, "layer1", ogr_geometry_type<point_type>::value, &spatial_reference);
			ogr_layer layer2(data_source, "layer2", ogr_geometry_type<point_type>::value, &spatial_reference);
			data_source.close();
			CPPUNIT_ASSERT( exists("test") );
			CPPUNIT_ASSERT( exists("test/layer1.shp") );
			CPPUNIT_ASSERT( exists("test/layer2.shp") );
			remove_all("test");
		}
		catch(CppUnit::Exception &e) { throw; }
		catch(std::exception &e) { CPPUNIT_FAIL( e.what() ); }
	}

	void test_ogr_point(const path &path, const point_type &expected_point1, const point_type &expected_point2) {
		ogr_data_source data_source(path, driver);
		CPPUNIT_ASSERT_EQUAL( 1, data_source->GetLayerCount() );
		ogr_layer layer(data_source->GetLayer(0));
		CPPUNIT_ASSERT( ogr_spatial_reference_equal(layer->GetSpatialRef(), &spatial_reference) );
#if GDAL_VERSION_MAJOR >= 2
		CPPUNIT_ASSERT_EQUAL( (GIntBig)2, layer->GetFeatureCount() );
#else
		CPPUNIT_ASSERT_EQUAL( 2, layer->GetFeatureCount() );
#endif
		ogr_feature feature1(layer->GetNextFeature());
		CPPUNIT_ASSERT_EQUAL( expected_point1, feature1.get_geometry<point_type>() );
		CPPUNIT_ASSERT_EQUAL( 1., feature1.get_field<double>("field") );
		ogr_feature feature2(layer->GetNextFeature());
		CPPUNIT_ASSERT_EQUAL( expected_point2, feature2.get_geometry<point_type>() );
		CPPUNIT_ASSERT_EQUAL( 2., feature2.get_field<double>(0) );
		data_source.close();
	}

	void test_ogr_point() {
		try {
			point_type point1(1, 1);
			point_type point2(2, 2);
			{
				ogr_data_source data_source("test_point.shp", driver, TRUE);
				data_source.delete_layers();
				ogr_layer layer = make_ogr_layer<point_type>(data_source, "layer", &spatial_reference);
				layer.create_field<double>("field");
				ogr_feature feature1(layer);
				feature1.set_geometry(point1);
				feature1.set_field("field", 1.);
				layer.create_feature(feature1);
				ogr_feature feature2(layer);
				feature2.set_geometry(point2);
				feature2.set_field(0, 2.);
				layer.create_feature(feature2);
				data_source.close();
				CPPUNIT_ASSERT( exists("test_point.shp") );
			}

			test_ogr_point("test_point.shp", point1, point2);

			point_type expected_point1(32.98940802777315184, 3.84116452878248538);
			point_type expected_point2(-45.7213722510897469, 47.61614393520889621);
			test_ogr_point("resources/geofis/ogr/point.shp", expected_point1, expected_point2);

			remove_all(".", regex(".*test_point\\..{3}.*"));
			}
		catch(CppUnit::Exception &e) { throw; }
		catch(std::exception &e) { CPPUNIT_FAIL( e.what() ); }
	}

	void test_ogr_polygon(const path &path, const polygon_type &expected_polygon1, const polygon_type &expected_polygon2) {
		ogr_data_source data_source(path, driver);
		CPPUNIT_ASSERT_EQUAL( 1, data_source->GetLayerCount() );
		ogr_layer layer(data_source->GetLayer(0));
		CPPUNIT_ASSERT( ogr_spatial_reference_equal(layer->GetSpatialRef(), &spatial_reference) );
#if GDAL_VERSION_MAJOR >= 2
		CPPUNIT_ASSERT_EQUAL( (GIntBig)2, layer->GetFeatureCount() );
#else
		CPPUNIT_ASSERT_EQUAL( 2, layer->GetFeatureCount() );
#endif
		ogr_feature feature1(layer->GetNextFeature());
		CPPUNIT_ASSERT_EQUAL( expected_polygon1, feature1.get_geometry<polygon_type>() );
		ogr_feature feature2(layer->GetNextFeature());
		CPPUNIT_ASSERT_EQUAL( expected_polygon2, feature2.get_geometry<polygon_type>() );
		data_source.close();
	}

	void test_ogr_polygon() {
		try {
			polygon_type polygon1 = make_polygon<kernel_type>(list_of(point_type(0, 0))(point_type(1, 0))(point_type(1, 1))(point_type(0, 1)));
			polygon_type polygon2 = make_polygon<kernel_type>(list_of(point_type(2, 0))(point_type(3, 0))(point_type(3, 1))(point_type(2, 1)));
			{
				ogr_data_source data_source("test_polygon.shp", driver, TRUE);
				data_source.delete_layers();
				ogr_layer layer = make_ogr_layer<polygon_type>(data_source, "layer", &spatial_reference);
				ogr_feature feature1(layer);
				feature1.set_geometry(polygon1);
				layer.create_feature(feature1);
				ogr_feature feature2(layer);
				feature2.set_geometry(polygon2);
				layer.create_feature(feature2);
				data_source.close();
				CPPUNIT_ASSERT( exists("test_polygon.shp") );
			}

			test_ogr_polygon("test_polygon.shp", polygon1, polygon2);

			polygon_type expected_polygon1 = make_polygon<kernel_type>(list_of(point_type(-11.22753030535277219, 23.90557365936026102))(point_type(9.43404951784874157, 18.88549672021922632))(point_type(13.97854022065009616, 37.59188868291317931))(point_type(-8.79675620850553486, 42.92902311294733408)));
			polygon_type expected_polygon2 = make_polygon<kernel_type>(list_of(point_type(-25.75933197128733454, 23.958416574509112))(point_type(-1.08169059677300083, 18.19853882328413874))(point_type(2.0888843121581786, 36.90493078597809529))(point_type(-29.56402186200475057, 42.66480853720305788)));
			test_ogr_polygon("resources/geofis/ogr/polygon.shp", expected_polygon1, expected_polygon2);

			remove_all(".", regex(".*test_polygon\\..{3}.*"));
		}
		catch(CppUnit::Exception &e) { throw; }
		catch(std::exception &e) { CPPUNIT_FAIL( e.what() ); }
	}

	void test_ogr_polygon_with_holes(const path &path, const polygon_with_holes_type &expected_polygon_with_holes) {
		ogr_data_source data_source(path, driver);
		CPPUNIT_ASSERT_EQUAL( 1, data_source->GetLayerCount() );
		ogr_layer layer(data_source->GetLayer(0));
		CPPUNIT_ASSERT( ogr_spatial_reference_equal(layer->GetSpatialRef(), &spatial_reference) );
#if GDAL_VERSION_MAJOR >= 2
		CPPUNIT_ASSERT_EQUAL( (GIntBig)1, layer->GetFeatureCount() );
#else
		CPPUNIT_ASSERT_EQUAL( 1, layer->GetFeatureCount() );
#endif
		ogr_feature feature(layer->GetNextFeature());
		CPPUNIT_ASSERT_EQUAL( expected_polygon_with_holes, feature.get_geometry<polygon_with_holes_type>() );
		data_source.close();
	}

	void test_ogr_polygon_with_holes() {
		try {
			polygon_type boundary = make_polygon<kernel_type>(list_of(point_type(0, 0))(point_type(5, 0))(point_type(5, 3))(point_type(0, 3)));
			polygon_type hole1 = make_polygon<kernel_type>(list_of(point_type(1, 1))(point_type(2, 1))(point_type(2, 2))(point_type(1, 2)));
			polygon_type hole2 = make_polygon<kernel_type>(list_of(point_type(3, 1))(point_type(4, 1))(point_type(4, 2))(point_type(3, 2)));
			polygon_with_holes_type polygon_with_holes = make_polygon_with_holes(boundary, list_of(hole1)(hole2));
			{
				ogr_data_source data_source("test_polygon_with_holes.shp", driver, TRUE);
				data_source.delete_layers();
				ogr_layer layer = make_ogr_layer<polygon_with_holes_type>(data_source, "layer", &spatial_reference);
				ogr_feature feature(layer);
				feature.set_geometry(polygon_with_holes);
				layer.create_feature(feature);
				data_source.close();
				CPPUNIT_ASSERT( exists("test_polygon_with_holes.shp") );
			}

			test_ogr_polygon_with_holes("test_polygon_with_holes.shp", polygon_with_holes);

			polygon_type expected_boundary = make_polygon<kernel_type>(list_of(point_type(-23.37368383190353072, 22.18865869285061621))(point_type(8.86980611440510458, 22.07172853892021536))(point_type(8.98673626833550898, 42.0083197840539313))(point_type(-23.40291637038613359, 41.92062216860612978)));
			polygon_type expected_hole1 = make_polygon<kernel_type>(list_of(point_type(-18.72571021317001083, 37.06802078049440752))(point_type(-18.63801259772220931, 29.02907269777919552))(point_type(-10.36520420714619206, 28.97060762081399332))(point_type(-10.24827405321579121, 37.09725331897700329)));
			polygon_type expected_hole2 = make_polygon<kernel_type>(list_of(point_type(-4.40176635669563865, 37.06802078049440752))(point_type(-4.46023143366084085, 28.47365446660977994))(point_type(3.63718172601957335, 28.35672431267937554))(point_type(3.87104203388037504, 37.06802078049440752)));
			polygon_with_holes_type expected_polygon_with_holes = make_polygon_with_holes(expected_boundary, list_of(expected_hole1)(expected_hole2));
			test_ogr_polygon_with_holes("resources/geofis/ogr/polygon_with_holes.shp", expected_polygon_with_holes);

			remove_all(".", regex(".*test_polygon_with_holes\\..{3}.*"));
		}
		catch(CppUnit::Exception &e) { throw; }
		catch(std::exception &e) { CPPUNIT_FAIL( e.what() ); }
	}
};

CPPUNIT_TEST_SUITE_REGISTRATION( ogr_tester );
