/**
 * GeoFIS Native Library
 *
 * Copyright (C) 2013 IRSTEA INRA SupAgro
 *
 * IDDN.FR.001.390016.000.R.P.2013.000.31235
 *
 * Auteurs/Authors:
 *     Jean-luc Lablée - IRSTEA
 *     Serge Guillaume - IRSTEA
 *     Brigitte Charnomordic - INRA
 *     Anne Tireau - INRA
 *     Bruno Tisseyre - SupAgro
 *
 * =========================================================================
 *
 * Ce logiciel est régi par la licence CeCILL soumise au droit français et
 * respectant les principes de diffusion des logiciels libres. Vous pouvez
 * utiliser, modifier et/ou redistribuer ce programme sous les conditions
 * de la licence CeCILL telle que diffusée par le CEA, le CNRS et l'INRIA
 * sur le site "http://www.cecill.info".
 *
 * En contrepartie de l'accessibilité au code source et des droits de copie,
 * de modification et de redistribution accordés par cette licence, il n'est
 * offert aux utilisateurs qu'une garantie limitée.  Pour les mêmes raisons,
 * seule une responsabilité restreinte pèse sur l'auteur du programme,  le
 * titulaire des droits patrimoniaux et les concédants successifs.
 *
 * A cet égard  l'attention de l'utilisateur est attirée sur les risques
 * associés au chargement,  à l'utilisation,  à la modification et/ou au
 * développement et à la reproduction du logiciel par l'utilisateur étant
 * donné sa spécificité de logiciel libre, qui peut le rendre complexe à
 * manipuler et qui le réserve donc à des développeurs et des professionnels
 * avertis possédant  des  connaissances  informatiques approfondies.  Les
 * utilisateurs sont donc invités à charger  et  tester  l'adéquation  du
 * logiciel à leurs besoins dans des conditions permettant d'assurer la
 * sécurité de leurs systèmes et ou de leurs données et, plus généralement,
 * à l'utiliser et l'exploiter dans les mêmes conditions de sécurité.
 *
 * Le fait que vous puissiez accéder à cet en-tête signifie que vous avez
 * pris connaissance de la licence CeCILL, et que vous en avez accepté les
 * termes.
 *
 * =========================================================================
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
#include <geofis/rcpp/process/zoning/map.hpp>
#include <gmock/gmock.h>
#include <CGAL/Cartesian.h>
#include <CGAL/Polygon_with_holes_2.h>
#include <boost/assign/list_of.hpp>
#include <boost/range/any_range.hpp>
#include <rcpp/rcpp_matchers.hpp>
#include <geofis/geometry/polygon.hpp>
#include <geofis/geometry/polygon_with_holes.hpp>

using namespace testing;
using namespace Rcpp;
using namespace std;
using namespace boost;
using namespace boost::assign;
using namespace CGAL;
using namespace geofis;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

typedef Cartesian<double> kernel_type;
typedef Point_2<kernel_type> point_type;
typedef Polygon_2<kernel_type> polygon_type;
typedef Polygon_with_holes_2<kernel_type> polygon_with_holes_type;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

namespace {

struct zone_mock {

	typedef polygon_with_holes_type geometry_type;

	MOCK_CONST_METHOD0( get_id, string () );
	MOCK_CONST_METHOD0( get_area, double () );
	MOCK_CONST_METHOD0( size, size_t () );
	MOCK_CONST_METHOD0( get_geometry, geometry_type () );
	MOCK_CONST_METHOD1( get_mean, double (size_t) );
	MOCK_CONST_METHOD1( get_standard_deviation, double (size_t) );
};

typedef any_range<const zone_mock, forward_traversal_tag> const_zone_range_type;

struct map_mock {

	MOCK_CONST_METHOD0( get_zones, const_zone_range_type () );
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

S4 make_expected_spatial_polygons() {
	string wkt = "GEOMETRYCOLLECTION (POLYGON ((0 0, 0 1, 1 1, 1 0 ,0 0)), POLYGON ((1 0, 1 1, 2 1, 2 0, 1 0)), POLYGON ((0 1, 0 2, 1 2, 1 1, 0 1)), POLYGON ((1 1, 1 2, 2 2, 2 1, 1 1)))";
	Function read_wkt("readWKT");
	S4 spatial_polygons = read_wkt(wkt);
	Function crs("CRS");
	spatial_polygons.slot("proj4string") = crs("+init=epsg:4326");
	return spatial_polygons;
}

DataFrame make_expected_data_frame() {
	CharacterVector ids = { "1", "2", "3", "4" };
	IntegerVector sizes = { 2, 3, 4, 5 };
	NumericVector areas = { 10, 11, 12, 13 };
	NumericVector a_means = { 1.5, 2.5, 3.5, 4.5 };
	NumericVector a_stds = { 11.5, 12.5, 13.5, 14.5 };
	NumericVector b_means = { 5.5, 6.5, 7.5, 8.5 };
	NumericVector b_stds = { 15.5, 16.5, 17.5, 18.5 };
	return DataFrame::create(Rcpp::Named("id") = ids,
			Rcpp::Named("size") = sizes,
			Rcpp::Named("area") = areas,
			Rcpp::Named("a_mean") = a_means,
			Rcpp::Named("a_std") = a_stds,
			Rcpp::Named("b_mean") = b_means,
			Rcpp::Named("b_std") = b_stds);
}

S4 make_expected_rcpp_map() {
	S4 Sr = make_expected_spatial_polygons();
	DataFrame data = make_expected_data_frame();
	Rcpp::Function spatial_polygons_data_frame("SpatialPolygonsDataFrame");
	return spatial_polygons_data_frame(Rcpp::Named("Sr") = Sr, Rcpp::Named("data") = data);
}

} // namespace

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

TEST(map_tester, test_rcpp_map) {

	polygon_type polygon1 = make_polygon<kernel_type>(list_of(point_type(0, 0))(point_type(1, 0))(point_type(1, 1))(point_type(0, 1)));
	polygon_type polygon2 = make_polygon<kernel_type>(list_of(point_type(1, 0))(point_type(2, 0))(point_type(2, 1))(point_type(1, 1)));
	polygon_type polygon3 = make_polygon<kernel_type>(list_of(point_type(0, 1))(point_type(1, 1))(point_type(1, 2))(point_type(0, 2)));
	polygon_type polygon4 = make_polygon<kernel_type>(list_of(point_type(1, 1))(point_type(2, 1))(point_type(2, 2))(point_type(1, 2)));

	zone_mock zone_mock1, zone_mock2, zone_mock3, zone_mock4;

	EXPECT_CALL( zone_mock1, get_id() ).WillRepeatedly(Return("1"));
	EXPECT_CALL( zone_mock1, size() ).WillRepeatedly(Return(2));
	EXPECT_CALL( zone_mock1, get_area() ).WillRepeatedly(Return(10));
	EXPECT_CALL( zone_mock1, get_mean(0) ).WillRepeatedly(Return(1.5));
	EXPECT_CALL( zone_mock1, get_mean(1) ).WillRepeatedly(Return(5.5));
	EXPECT_CALL( zone_mock1, get_standard_deviation(0) ).WillRepeatedly(Return(11.5));
	EXPECT_CALL( zone_mock1, get_standard_deviation(1) ).WillRepeatedly(Return(15.5));
	EXPECT_CALL( zone_mock1, get_geometry() ).WillRepeatedly(Return(make_polygon_with_holes<kernel_type>(polygon1)));

	EXPECT_CALL( zone_mock2, get_id() ).WillRepeatedly(Return("2"));
	EXPECT_CALL( zone_mock2, size() ).WillRepeatedly(Return(3));
	EXPECT_CALL( zone_mock2, get_area() ).WillRepeatedly(Return(11));
	EXPECT_CALL( zone_mock2, get_mean(0) ).WillRepeatedly(Return(2.5));
	EXPECT_CALL( zone_mock2, get_mean(1) ).WillRepeatedly(Return(6.5));
	EXPECT_CALL( zone_mock2, get_standard_deviation(0) ).WillRepeatedly(Return(12.5));
	EXPECT_CALL( zone_mock2, get_standard_deviation(1) ).WillRepeatedly(Return(16.5));
	EXPECT_CALL( zone_mock2, get_geometry() ).WillRepeatedly(Return(make_polygon_with_holes<kernel_type>(polygon2)));

	EXPECT_CALL( zone_mock3, get_id() ).WillRepeatedly(Return("3"));
	EXPECT_CALL( zone_mock3, size() ).WillRepeatedly(Return(4));
	EXPECT_CALL( zone_mock3, get_area() ).WillRepeatedly(Return(12));
	EXPECT_CALL( zone_mock3, get_mean(0) ).WillRepeatedly(Return(3.5));
	EXPECT_CALL( zone_mock3, get_mean(1) ).WillRepeatedly(Return(7.5));
	EXPECT_CALL( zone_mock3, get_standard_deviation(0) ).WillRepeatedly(Return(13.5));
	EXPECT_CALL( zone_mock3, get_standard_deviation(1) ).WillRepeatedly(Return(17.5));
	EXPECT_CALL( zone_mock3, get_geometry() ).WillRepeatedly(Return(make_polygon_with_holes<kernel_type>(polygon3)));

	EXPECT_CALL( zone_mock4, get_id() ).WillRepeatedly(Return("4"));
	EXPECT_CALL( zone_mock4, size() ).WillRepeatedly(Return(5));
	EXPECT_CALL( zone_mock4, get_area() ).WillRepeatedly(Return(13));
	EXPECT_CALL( zone_mock4, get_mean(0) ).WillRepeatedly(Return(4.5));
	EXPECT_CALL( zone_mock4, get_mean(1) ).WillRepeatedly(Return(8.5));
	EXPECT_CALL( zone_mock4, get_standard_deviation(0) ).WillRepeatedly(Return(14.5));
	EXPECT_CALL( zone_mock4, get_standard_deviation(1) ).WillRepeatedly(Return(18.5));
	EXPECT_CALL( zone_mock4, get_geometry() ).WillRepeatedly(Return(make_polygon_with_holes<kernel_type>(polygon4)));

	map_mock map_mock;
	EXPECT_CALL( map_mock, get_zones() ).WillRepeatedly(Return(cref_list_of<4>(zone_mock1)(zone_mock2)(zone_mock3)(zone_mock4)));

	Function crs_function("CRS");
	S4 crs = crs_function("+init=epsg:4326");

	CharacterVector names = { "a", "b" };

	S4 rcpp_map = make_rcpp_map(map_mock, crs, names);

	S4 expected_rcpp_map = make_expected_rcpp_map();

	ASSERT_THAT(rcpp_map, Identical(expected_rcpp_map));
}
