/**
 * GeoFIS Native Library
 *
 * Copyright (C) 2013 IRSTEA INRA SupAgro
 *
 * IDDN.FR.001.390016.000.R.P.2013.000.31235
 *
 * Auteurs/Authors:
 *     Jean-luc Lablée - IRSTEA
 *     Serge Guillaume - IRSTEA
 *     Brigitte Charnomordic - INRA
 *     Anne Tireau - INRA
 *     Bruno Tisseyre - SupAgro
 *
 * =========================================================================
 *
 * Ce logiciel est régi par la licence CeCILL soumise au droit français et
 * respectant les principes de diffusion des logiciels libres. Vous pouvez
 * utiliser, modifier et/ou redistribuer ce programme sous les conditions
 * de la licence CeCILL telle que diffusée par le CEA, le CNRS et l'INRIA
 * sur le site "http://www.cecill.info".
 *
 * En contrepartie de l'accessibilité au code source et des droits de copie,
 * de modification et de redistribution accordés par cette licence, il n'est
 * offert aux utilisateurs qu'une garantie limitée.  Pour les mêmes raisons,
 * seule une responsabilité restreinte pèse sur l'auteur du programme,  le
 * titulaire des droits patrimoniaux et les concédants successifs.
 *
 * A cet égard  l'attention de l'utilisateur est attirée sur les risques
 * associés au chargement,  à l'utilisation,  à la modification et/ou au
 * développement et à la reproduction du logiciel par l'utilisateur étant
 * donné sa spécificité de logiciel libre, qui peut le rendre complexe à
 * manipuler et qui le réserve donc à des développeurs et des professionnels
 * avertis possédant  des  connaissances  informatiques approfondies.  Les
 * utilisateurs sont donc invités à charger  et  tester  l'adéquation  du
 * logiciel à leurs besoins dans des conditions permettant d'assurer la
 * sécurité de leurs systèmes et ou de leurs données et, plus généralement,
 * à l'utiliser et l'exploiter dans les mêmes conditions de sécurité.
 *
 * Le fait que vous puissiez accéder à cet en-tête signifie que vous avez
 * pris connaissance de la licence CeCILL, et que vous en avez accepté les
 * termes.
 *
 * =========================================================================
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
#include <geofis/rcpp/process/zoning/neighborhood/neighborhood_map.hpp>
#include <gmock/gmock.h>
#include <CGAL/Cartesian.h>
#include <CGAL/Point_2.h>
#include <boost/range/any_range.hpp>
#include <boost/assign/list_of.hpp>
#include <rcpp/rcpp_matchers.hpp>

#include <boost/assign/std/vector.hpp>

using namespace testing;
using namespace Rcpp;
using namespace std;
using namespace boost;
using namespace boost::assign;
using namespace CGAL;
using namespace geofis;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

typedef Cartesian<double> kernel_type;
typedef Point_2<kernel_type> point_type;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

namespace {

struct feature_mock {

	MOCK_CONST_METHOD0( get_geometry, point_type () );
};

struct zone_mock {

	MOCK_CONST_METHOD1( get_feature, const feature_mock & (size_t) );
};

struct zone_neighbor_mock {

	MOCK_CONST_METHOD0( get_zone1, const zone_mock & () );
	MOCK_CONST_METHOD0( get_zone2, const zone_mock & () );
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

S4 make_expected_spatial_lines() {
	string wkt = "GEOMETRYCOLLECTION (LINESTRING (0 1, 1 1), LINESTRING (0 1, 0 0), LINESTRING (1 1, 1 0), LINESTRING (0 0, 1 0), LINESTRING (0 1, 1 0), LINESTRING (1 1, 0 0))";
	Function read_wkt("readWKT");
	S4 spatial_lines = read_wkt(wkt);
	Function crs("CRS");
	spatial_lines.slot("proj4string") = crs("+init=epsg:4326");
	return spatial_lines;
}

DataFrame make_expected_data_frame() {
	LogicalVector filtered = { false, false, false, false, true, true };
	return DataFrame::create(Rcpp::Named("filtered") = filtered);
}

S4 make_expected_rcpp_neighborhood_map() {
	S4 sl = make_expected_spatial_lines();
	DataFrame data = make_expected_data_frame();
	Rcpp::Function spatial_lines_data_frame("SpatialLinesDataFrame");
	return spatial_lines_data_frame(Rcpp::Named("sl") = sl, Rcpp::Named("data") = data);
}

} // namespace

typedef any_range<const zone_neighbor_mock, forward_traversal_tag> zone_neighbor_range_type;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

TEST(neighborhood_map_tester, test_neighborhood_map) {

	point_type point1(0, 1), point2(1, 1), point3(0, 0), point4(1, 0);

	feature_mock feature1, feature2, feature3, feature4;
	EXPECT_CALL( feature1, get_geometry() ).WillRepeatedly(Return(point1));
	EXPECT_CALL( feature2, get_geometry() ).WillRepeatedly(Return(point2));
	EXPECT_CALL( feature3, get_geometry() ).WillRepeatedly(Return(point3));
	EXPECT_CALL( feature4, get_geometry() ).WillRepeatedly(Return(point4));

	zone_mock zone1, zone2, zone3, zone4;
	EXPECT_CALL( zone1, get_feature(0) ).WillRepeatedly(ReturnRef(feature1));
	EXPECT_CALL( zone2, get_feature(0) ).WillRepeatedly(ReturnRef(feature2));
	EXPECT_CALL( zone3, get_feature(0) ).WillRepeatedly(ReturnRef(feature3));
	EXPECT_CALL( zone4, get_feature(0) ).WillRepeatedly(ReturnRef(feature4));

	zone_neighbor_mock zone_neighbor1, zone_neighbor2, zone_neighbor3, zone_neighbor4, zone_neighbor5, zone_neighbor6;
	EXPECT_CALL( zone_neighbor1, get_zone1() ).WillRepeatedly(ReturnRef(zone1));
	EXPECT_CALL( zone_neighbor1, get_zone2() ).WillRepeatedly(ReturnRef(zone2));
	EXPECT_CALL( zone_neighbor2, get_zone1() ).WillRepeatedly(ReturnRef(zone1));
	EXPECT_CALL( zone_neighbor2, get_zone2() ).WillRepeatedly(ReturnRef(zone3));
	EXPECT_CALL( zone_neighbor3, get_zone1() ).WillRepeatedly(ReturnRef(zone2));
	EXPECT_CALL( zone_neighbor3, get_zone2() ).WillRepeatedly(ReturnRef(zone4));
	EXPECT_CALL( zone_neighbor4, get_zone1() ).WillRepeatedly(ReturnRef(zone3));
	EXPECT_CALL( zone_neighbor4, get_zone2() ).WillRepeatedly(ReturnRef(zone4));
	EXPECT_CALL( zone_neighbor5, get_zone1() ).WillRepeatedly(ReturnRef(zone1));
	EXPECT_CALL( zone_neighbor5, get_zone2() ).WillRepeatedly(ReturnRef(zone4));
	EXPECT_CALL( zone_neighbor6, get_zone1() ).WillRepeatedly(ReturnRef(zone2));
	EXPECT_CALL( zone_neighbor6, get_zone2() ).WillRepeatedly(ReturnRef(zone3));

	Function crs_function("CRS");
	S4 crs = crs_function("+init=epsg:4326");

	auto neighbor_range = cref_list_of<4>(zone_neighbor1)(zone_neighbor2)(zone_neighbor3)(zone_neighbor4);
	auto filtered_neighbor_range = cref_list_of<2>(zone_neighbor5)(zone_neighbor6);

	S4 rcpp_neighborhood_map = make_rcpp_neighborhood_map(zone_neighbor_range_type(neighbor_range), zone_neighbor_range_type(filtered_neighbor_range), crs);

	S4 expected_rcpp_neighborhood_map = make_expected_rcpp_neighborhood_map();

	ASSERT_THAT(rcpp_neighborhood_map, Identical(expected_rcpp_neighborhood_map));
}
