/**
 * Util Native Library
 *
 * Copyright (C) 2014 IRSTEA
 *
 * Auteur/Author:
 *     Jean-luc Lablée - http://itap.irstea.fr/?page_id=2277
 *
 * =========================================================================
 *
 * Ce logiciel est régi par la licence CeCILL-C soumise au droit français et
 * respectant les principes de diffusion des logiciels libres. Vous pouvez
 * utiliser, modifier et/ou redistribuer ce programme sous les conditions
 * de la licence CeCILL-C telle que diffusée par le CEA, le CNRS et l'INRIA
 * sur le site "http://www.cecill.info".
 *
 * En contrepartie de l'accessibilité au code source et des droits de copie,
 * de modification et de redistribution accordés par cette licence, il n'est
 * offert aux utilisateurs qu'une garantie limitée.  Pour les mêmes raisons,
 * seule une responsabilité restreinte pèse sur l'auteur du programme,  le
 * titulaire des droits patrimoniaux et les concédants successifs.
 *
 * A cet égard  l'attention de l'utilisateur est attirée sur les risques
 * associés au chargement,  à l'utilisation,  à la modification et/ou au
 * développement et à la reproduction du logiciel par l'utilisateur étant
 * donné sa spécificité de logiciel libre, qui peut le rendre complexe à
 * manipuler et qui le réserve donc à des développeurs et des professionnels
 * avertis possédant  des  connaissances  informatiques approfondies.  Les
 * utilisateurs sont donc invités à charger  et  tester  l'adéquation  du
 * logiciel à leurs besoins dans des conditions permettant d'assurer la
 * sécurité de leurs systèmes et ou de leurs données et, plus généralement,
 * à l'utiliser et l'exploiter dans les mêmes conditions de sécurité.
 *
 * Le fait que vous puissiez accéder à cet en-tête signifie que vous avez
 * pris connaissance de la licence CeCILL-C, et que vous en avez accepté les
 * termes.
 *
 * =========================================================================
 *
 * This software is governed by the CeCILL-C license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-C
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
 */
#include <util/convert/convert.hpp>
#include <cassert>
#include <cstdlib>
#include <cstring>
#include <iconv.h>
#include <stdexcept>
#include <cerrno>
#include <boost/format.hpp>
#include <boost/type_traits/integral_constant.hpp>
#include <localcharset.h>

namespace util {

extern "C" {
    typedef size_t (*gnu_iconv_ptr_type)(iconv_t d, const char **input, size_t *input_size, char **output, size_t *output_size);
    typedef size_t (*posix_iconv_ptr_type)(iconv_t d, char **input, size_t *input_size, char **output, size_t *output_size);
}

inline size_t do_iconv(gnu_iconv_ptr_type iconv_ptr, iconv_t d, const char **input, size_t *input_size, char **output, size_t *output_size) {
   return iconv_ptr(d, input, input_size, output, output_size);
}

inline size_t do_iconv(posix_iconv_ptr_type iconv_ptr, iconv_t d, const char **input, size_t *input_size, char **output, size_t *output_size) {
    return iconv_ptr(d, const_cast<char **>(input), input_size, output, output_size);
}

inline size_t call_iconv(iconv_t d, const char **input, size_t *input_size, char **output, size_t *output_size) {
    return do_iconv(iconv, d, input, input_size, output, output_size);
}

static auto_malloc<char> iconv_converter(const char *input, size_t input_size, const char *to_charset, const char *from_charset) {
	assert(input);
	assert(from_charset);
	assert(to_charset);

	size_t inleft, outleft, converted = 0;
	char *outbuf;
	const char *inbuf;
	size_t outlen;

	struct iconv_raii {

		iconv_raii(const char* to_charset, const char* from_charset) {
			converter = iconv_open(to_charset, from_charset);
			if(converter == (iconv_t)(-1))
				throw std::runtime_error(boost::str(boost::format("The conversion from \"%1%\" to \"%2%\" is not supported by the implementation") % from_charset % to_charset));
		}
		~iconv_raii() { iconv_close(converter); }

		iconv_t converter;

	} raii_converter(to_charset, from_charset);

	inleft = input_size;
	inbuf = input;

	// we'll start off allocating an output buffer which is the same size as our input buffer.
	outlen = inleft;

	// we allocate 4 bytes more than what we need for nul-termination...
	auto_malloc<char> output(outlen + 4);

	while(1) {

		errno = 0;
		outbuf = output.get() + converted;
		outleft = outlen - converted;

		if(call_iconv(raii_converter.converter, &inbuf, &inleft, &outbuf, &outleft) == (size_t)-1) {
			if (errno != E2BIG)
				// An invalid multibyte sequence has been encountered in the input.
				// we can't really recover from this.
				throw std::runtime_error("invalid conversion");

			else {
				// E2BIG   There is not sufficient room at *outbuf.
				// We just need to grow our outbuffer and try again.
				converted = outbuf - output.get();
				outlen += inleft * 2;
				output.realloc(outlen + 4);
			}
		} else {

			// Note: not all charsets can be nul-terminated with a single nul byte.
			// UCS2, for example, needs 2 nul bytes and UCS4 needs 4.
			// I hope that 4 nul bytes is enough to terminate all multibyte charsets?
			memset(outbuf, 0, 4); // nul-terminate the string
			return boost::move(output);
		}
	}
}

template <> convert_result<char> convert<char, char>(const char *input, const char* to_charset, const char* from_charset) {
	return convert_result<char>(iconv_converter(input, strlen(input), to_charset, from_charset));
}

template <> convert_result<wchar_t> convert<wchar_t, char>(const char *input, const char* to_charset, const char* from_charset) {
	return convert_result<wchar_t>(iconv_converter(input, strlen(input), to_charset, from_charset));
}

template <> convert_result<char> convert<char, wchar_t>(const wchar_t *input, const char* to_charset, const char* from_charset) {
	return convert_result<char>(iconv_converter((const char *)input, wcslen(input) * sizeof(wchar_t), to_charset, from_charset));
}

convert_result<char> convert_iso88591_to_utf8(const char *input) { return convert<char, char>(input, "UTF-8", "ISO-8859-1"); }
convert_result<wchar_t> convert_iso88591_to_unicode(const char *input) { return convert<wchar_t, char>(input, "UCS-2-INTERNAL", "ISO-8859-1"); }

convert_result<char> convert_utf8_to_iso88591(const char *input) { return convert<char, char>(input, "ISO-8859-1", "UTF-8"); }
convert_result<wchar_t> convert_utf8_to_unicode(const char *input) { return convert<wchar_t, char>(input, "UCS-2-INTERNAL", "UTF-8"); }
convert_result<char> convert_utf8_to_local_charset(const char *input) { return convert<char, char>(input, locale_charset(), "UTF-8"); }

convert_result<char> convert_unicode_to_iso88591(const wchar_t *input) { return convert<char, wchar_t>(input, "ISO-8859-1", "UCS-2-INTERNAL"); }
convert_result<char> convert_unicode_to_utf8(const wchar_t *input) { return convert<char, wchar_t>(input, "UTF-8", "UCS-2-INTERNAL"); }

convert_result<char> convert_local_charset_to_utf8(const char *input) { return convert<char, char>(input, "UTF-8", locale_charset()); }

} // namespace util

